#include "Application.h"
#include <fstream>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include "BearingCalculator.h"
#include "DistanceParser.h"
#include "AngleParser.h"
#define	THISCLASS Application

int runapp(int argc, char * argv[]) {
	Application app;
	return app.Run(argc, argv);
}

int main(int argc, char * argv[]) {
	// Run the application
	int ret = runapp(argc, argv);

	// Return with the exit code
	return ret;
}

THISCLASS::Application():
	mBearingCalculator(), mHelpMode(false), mAngleFormat(AngleDisplay::sFormatDegMinSec),
	mDistanceFormat(DistanceDisplay::sFormatMeter) {
}

THISCLASS::~Application() {
}

int THISCLASS::Run(int argc, char * argv[]) {
	// Process arguments
	int argi = 1;
	int argnw = 0;
	static const int argmax = 10;
	std::string option = "";
	std::string args[argmax];
	while (argi < argc) {
		std::string arg = argv[argi];

		if (arg.length() > 1 && arg[0] == '-' && arg[1] == '-') {
			if (option.length()) {
				if (! SetOption(option, args, argnw)) return 2;
			}
			argnw = 0;
			option = arg;
		} else {
			if (option.length() == 0) {
				UnknownOption(arg);
				return 1;
			}
			if (argnw >= argmax) {
				TooManyArguments(option);
				return 1;
			}
			args[argnw] = arg;
			argnw++;
		}
		argi++;
	}

	if (option.length()) {
		if (! SetOption(option, args, argnw)) return 2;
	}

	// Show help
	if (mHelpMode) {
		Help();
		return 0;
	}

	// Calculate
	mBearingCalculator.Calculate();

	// Show Results
	if (ShowResults()) return 0;
	return 3;
}

bool THISCLASS::ShowResults() {
	if (! mBearingCalculator.mResultsAvailable) {
		std::cerr << "The angles could not be calculated." << std::endl;
		return false;
	}

	Angle a180;
	a180.SetDeg(-180);

	std::cout << "=== MY POSITION AND ORIENTATION ===" << std::endl;
	std::cout << "Latitude: " << AngleDisplay(mBearingCalculator.mMyPosition.mLatitude.Normalize(a180), mAngleFormat) << std::endl;
	std::cout << "Longitude: " << AngleDisplay(mBearingCalculator.mMyPosition.mLongitude.Normalize(a180), mAngleFormat) << std::endl;
	std::cout << "Altitude: " << DistanceDisplay(mBearingCalculator.mMyPosition.mAltitude, mDistanceFormat) << std::endl;

	std::cout << "Heading: " << AngleDisplay(mBearingCalculator.mMyHeading.Normalize(0), mAngleFormat) << std::endl;
	std::cout << "Pitch: " << AngleDisplay(mBearingCalculator.mMyPitch.Normalize(a180), mAngleFormat) << std::endl;
	std::cout << "Roll: " << AngleDisplay(mBearingCalculator.mMyRoll.Normalize(a180), mAngleFormat) << std::endl;

	std::cout << "Orientation matrix: ";
	std::cout << "((" << mBearingCalculator.mMyOrientation.a[0][0] << ", " << mBearingCalculator.mMyOrientation.a[0][1] << ", " << mBearingCalculator.mMyOrientation.a[0][2] << ")";
	std::cout << ", (" << mBearingCalculator.mMyOrientation.a[1][0] << ", " << mBearingCalculator.mMyOrientation.a[1][1] << ", " << mBearingCalculator.mMyOrientation.a[1][2] << ")";
	std::cout << ", (" << mBearingCalculator.mMyOrientation.a[2][0] << ", " << mBearingCalculator.mMyOrientation.a[2][1] << ", " << mBearingCalculator.mMyOrientation.a[2][2] << "))" << std::endl;

	std::cout << std::endl;
	std::cout << "=== TARGET POSITION ===" << std::endl;
	std::cout << "Latitude: " << AngleDisplay(mBearingCalculator.mTargetPosition.mLatitude.Normalize(a180), mAngleFormat) << std::endl;
	std::cout << "Longitude: " << AngleDisplay(mBearingCalculator.mTargetPosition.mLongitude.Normalize(a180), mAngleFormat) << std::endl;
	std::cout << "Altitude: " << DistanceDisplay(mBearingCalculator.mTargetPosition.mAltitude, mDistanceFormat) << std::endl;

	std::cout << std::endl;
	std::cout << "=== RESULT ===" << std::endl;
	std::cout << "Horizontal angle: " << AngleDisplay(mBearingCalculator.mHorizontalAngle.Normalize(0), mAngleFormat) << std::endl;
	std::cout << "Vertical angle: " << AngleDisplay(mBearingCalculator.mVerticalAngle.Normalize(a180), mAngleFormat) << std::endl;
	std::cout << "Distance: " << DistanceDisplay(mBearingCalculator.mDistance, mDistanceFormat) << std::endl;
	return true;
}

bool THISCLASS::SetOption(const std::string &option, const std::string arg[], int argcc) {
	//std::cerr << "Option '" << option << "' with first argument '" << arg[0] << "'" << std::endl;
	if (option == "--help" || option == "--h") {
		mHelpMode = true;
	} else if (option == "--angle-format") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		if (arg[0] == "rad") {
			mAngleFormat = AngleDisplay::sFormatRad;
		} else if (arg[0] == "deg") {
			mAngleFormat = AngleDisplay::sFormatDeg;
		} else if (arg[0] == "degminsec") {
			mAngleFormat = AngleDisplay::sFormatDegMinSec;
		} else if (arg[0] == "grad") {
			mAngleFormat = AngleDisplay::sFormatGrad;
		} else {
			std::cerr << "Unknown angle format: " << arg[0] << std::endl;
			return false;
		}
	} else if (option == "--distance-format") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		if (arg[0] == "plain") {
			mDistanceFormat = DistanceDisplay::sFormatPlain;
		} else if (arg[0] == "meter") {
			mDistanceFormat = DistanceDisplay::sFormatMeter;
		} else if (arg[0] == "kilometer") {
			mDistanceFormat = DistanceDisplay::sFormatKilometer;
		} else {
			std::cerr << "Unknown distance format: " << arg[0] << std::endl;
			return false;
		}
	} else if (option == "--my-position") {
		if (! CheckArgumentCount(option, argcc, 2, 3)) return false;
		mBearingCalculator.mMyPosition.mLatitude = AngleParser(arg[0], true, false).GetValue();
		mBearingCalculator.mMyPosition.mLongitude = AngleParser(arg[1], false, true).GetValue();
		mBearingCalculator.mMyPosition.mAltitude = (argcc > 2 ? DistanceParser(arg[2]).GetValue() : 0);
	} else if (option == "--my-latitude") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mMyPosition.mLatitude = AngleParser(arg[0]).GetValue();
	} else if (option == "--my-longitude") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mMyPosition.mLongitude = AngleParser(arg[0]).GetValue();
	} else if (option == "--my-altitude") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mMyPosition.mAltitude = DistanceParser(arg[0]).GetValue();
	} else if (option == "--target-position") {
		if (! CheckArgumentCount(option, argcc, 2, 3)) return false;
		mBearingCalculator.mTargetPosition.mLatitude = AngleParser(arg[0], true, false).GetValue();
		mBearingCalculator.mTargetPosition.mLongitude = AngleParser(arg[1], false, true).GetValue();
		mBearingCalculator.mTargetPosition.mAltitude = (argcc > 2 ? DistanceParser(arg[2]).GetValue() : 0);
	} else if (option == "--target-latitude") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mTargetPosition.mLatitude = AngleParser(arg[0], true, false).GetValue();
	} else if (option == "--target-longitude") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mTargetPosition.mLongitude = AngleParser(arg[0], false, true).GetValue();
	} else if (option == "--target-altitude") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mTargetPosition.mAltitude = DistanceParser(arg[0]).GetValue();
	} else if (option == "--my-heading") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mMyHeading = AngleParser(arg[0]).GetValue();
	} else if (option == "--my-pitch") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mMyPitch = AngleParser(arg[0]).GetValue();
	} else if (option == "--my-roll") {
		if (! CheckArgumentCount(option, argcc, 1, 1)) return false;
		mBearingCalculator.mMyRoll = AngleParser(arg[0]).GetValue();
	} else if (option == "--my-orientation") {
		if (! CheckArgumentCount(option, argcc, 9, 9)) return false;
		int i = 0;
		for (int y = 0; y < 3; y++) {
			for (int x = 0; x < 3; x++) {
				std::istringstream str(arg[i]);
				str >> mBearingCalculator.mMyOrientation.a[y][x];
				i++;
			}
		}
	} else {
		return UnknownOption(option);
	}

	return true;
}

void THISCLASS::Help() {
	std::cout << "GPS Bearing and Distance Calculator" << std::endl;
	std::cout << "Usage: gps-bearing [OPTIONS]" << std::endl;
	std::cout << std::endl;
	std::cout << "Options:" << std::endl;
	std::cout << "  --help                            Shows this help" << std::endl;
	std::cout << "  --angle-format rad                Displays angles in radians" << std::endl;
	std::cout << "  --angle-format deg                Displays angles in degrees" << std::endl;
	std::cout << "  --angle-format degminsec          Displays angles in degrees, minutes and seconds" << std::endl;
	std::cout << "  --angle-format grad               Displays angles in grad" << std::endl;
	std::cout << "  --distance-format plain           Displays distances in meters (without unit)" << std::endl;
	std::cout << "  --distance-format meter           Displays distances in meters" << std::endl;
	std::cout << "  --distance-format kilometer       Displays distances in kilometers" << std::endl;
	std::cout << "  --my-position LAT LONG [ALT]      My position" << std::endl;
	std::cout << "  --my-latitude LAT                 My latitude" << std::endl;
	std::cout << "  --my-longitude LONG               My longitude" << std::endl;
	std::cout << "  --my-altitude ALT                 My altitude" << std::endl;
	std::cout << "  --my-orientation MATRIX           9 values (3x3 matrix) that describes how the antenna is mounted" << std::endl;
	std::cout << "  --my-heading ANGLE                My heading" << std::endl;
	std::cout << "  --my-pitch ANGLE                  My pitch" << std::endl;
	std::cout << "  --my-roll ANGLE                   My roll" << std::endl;
	std::cout << "  --target-position LAT LONG [ALT]  Target position" << std::endl;
	std::cout << "  --target-latitude LAT             Target latitude" << std::endl;
	std::cout << "  --target-longitude LONG           Target longitude" << std::endl;
	std::cout << "  --target-altitude ALT             Target altitude" << std::endl;
	std::cout << std::endl;
	std::cout << "Please report bugs to Thomas Lochmatter, thomas.lochmatter@viereck.ch" << std::endl;
}

bool THISCLASS::UnknownOption(const std::string &option) {
	std::cerr << "Unknown option: " << option << std::endl;
	std::cerr << "Type 'gpsangle --help' to get a list of options." << std::endl;
	return false;
}

bool THISCLASS::TooManyArguments(const std::string &option) {
	std::cerr << "Too many arguments for " << option << std::endl;
	std::cerr << "Type 'gps-bearing --help' to get a list of options." << std::endl;
	return false;
}

bool THISCLASS::CheckArgumentCount(const std::string &option, int argcc, int min, int max) {
	if (argcc < min) {
		std::cerr << "Not enough arguments for " << option << std::endl;
		std::cerr << "Type 'gps-bearing --help' to get a list of options." << std::endl;
		return false;
	}

	if (argcc > max) {
		std::cerr << "Too many arguments for " << option << std::endl;
		std::cerr << "Type 'gps-bearing --help' to get a list of options." << std::endl;
		return false;
	}

	return true;
}
