#include "RotationMatrix.h"
#define THISCLASS RotationMatrix

THISCLASS::RotationMatrix() {
	SetIdentity();
}

THISCLASS::RotationMatrix(double a11, double a12, double a13, double a21, double a22, double a23, double a31, double a32, double a33) {
	a[0][0] = a11;
	a[0][1] = a12;
	a[0][2] = a13;
	a[1][0] = a21;
	a[1][1] = a22;
	a[1][2] = a23;
	a[2][0] = a31;
	a[2][1] = a32;
	a[2][2] = a33;
}

void THISCLASS::SetIdentity() {
	a[0][0] = 1;
	a[0][1] = 0;
	a[0][2] = 0;
	a[1][0] = 0;
	a[1][1] = 1;
	a[1][2] = 0;
	a[2][0] = 0;
	a[2][1] = 0;
	a[2][2] = 1;
}

void THISCLASS::SetVectors(const Vector3d &c1, const Vector3d &c2, const Vector3d &c3) {
	a[0][0] = c1.x;
	a[0][1] = c1.y;
	a[0][2] = c1.z;
	a[1][0] = c2.x;
	a[1][1] = c2.y;
	a[1][2] = c2.z;
	a[2][0] = c3.x;
	a[2][1] = c3.y;
	a[2][2] = c3.z;
}

void THISCLASS::SetGyro(Angle pitch, Angle roll, Angle yaw) {
	double cospitch = cos(pitch.GetRad());
	double sinpitch = sin(pitch.GetRad());
	double cosroll = cos(roll.GetRad());
	double sinroll = sin(roll.GetRad());
	double cosyaw = cos(yaw.GetRad());
	double sinyaw = sin(yaw.GetRad());

	// X
	a[0][0] = cospitch * cosyaw;
	a[1][0] = -cospitch * sinyaw;
	a[2][0] = sinpitch;

	// Y
	a[0][1] = sinroll * sinpitch * cosyaw + cosroll * sinyaw;
	a[1][1] = -sinroll * sinpitch * sinyaw + cosroll * cosyaw;
	a[2][1] = -cospitch * sinroll;

	// Z
	a[0][2] = -cosroll * sinpitch * cosyaw + sinroll * sinyaw;
	a[1][2] = cosroll * sinpitch * sinyaw + sinroll * cosyaw;
	a[2][2] = cospitch * cosroll;
}

void THISCLASS::SetSpheric(Angle latitude, Angle longitude) {
	double coslat = cos(latitude.GetRad());
	double sinlat = sin(latitude.GetRad());
	double coslong = cos(longitude.GetRad());
	double sinlong = sin(longitude.GetRad());

	// X: 90-lat, long+180
	a[0][0] = -sinlat * coslong;
	a[1][0] = -sinlat * sinlong;
	a[2][0] = coslat;

	// Y: 0, long-90
	a[0][1] = sinlong;
	a[1][1] = -coslong;
	a[2][1] = 0;

	// Z: lat, long
	a[0][2] = coslat * coslong;
	a[1][2] = coslat * sinlong;
	a[2][2] = sinlat;
}

double THISCLASS::SubDeterminant(int omitrow, int omitcol) const {
	int row1, row2;
	int col1, col2;

	if (omitrow == 0) {
		row1 = 1;
		row2 = 2;
	} else if (omitrow == 1) {
		row1 = 0;
		row2 = 2;
	} else {
		row1 = 0;
		row2 = 1;
	}

	if (omitcol == 0) {
		col1 = 1;
		col2 = 2;
	} else if (omitcol == 1) {
		col1 = 0;
		col2 = 2;
	} else {
		col1 = 0;
		col2 = 1;
	}

	return SubDeterminant(row1, row2, col1, col2);
}

double THISCLASS::SubDeterminant(int row1, int row2, int col1, int col2) const {
	return a[row1][col1] * a[row2][col2] - a[row2][col1] * a[row1][col2];
}

double THISCLASS::GeneralDeterminant() const {
	return a[0][0] * SubDeterminant(1, 2, 1, 2) - a[1][0] * SubDeterminant(0, 2, 1, 2) + a[2][0] * SubDeterminant(0, 1, 1, 2);
}

RotationMatrix THISCLASS::GeneralInverse() const {
	RotationMatrix inv;
	double det = GeneralDeterminant();
	double sign = 1;
	for (int x = 0; x < 3; x++) {
		for (int y = 0; y < 3; y++) {
			inv.a[y][x] = sign * SubDeterminant(x, y) / det;
			sign = -sign;
		}
	}
	return inv;
}

RotationMatrix THISCLASS::Transpose() const {
	RotationMatrix inv;
	for (int x = 0; x < 3; x++) {
		for (int y = 0; y < 3; y++) {
			inv.a[y][x] = a[x][y];
		}
	}
	return inv;
}

Vector3d THISCLASS::Multiply(const Vector3d &v) const {
	Vector3d res;
	res.x = a[0][0] * v.x + a[0][1] * v.y + a[0][2] * v.z;
	res.y = a[1][0] * v.x + a[1][1] * v.y + a[1][2] * v.z;
	res.z = a[2][0] * v.x + a[2][1] * v.y + a[2][2] * v.z;
	return res;
}

RotationMatrix THISCLASS::Multiply(const RotationMatrix &m) const {
	RotationMatrix res;

	// First column
	res.a[0][0] = a[0][0] * m.a[0][0] + a[0][1] * m.a[1][0] + a[0][2] * m.a[2][0];
	res.a[1][0] = a[1][0] * m.a[0][0] + a[1][1] * m.a[1][0] + a[1][2] * m.a[2][0];
	res.a[2][0] = a[2][0] * m.a[0][0] + a[2][1] * m.a[1][0] + a[2][2] * m.a[2][0];

	// Second column
	res.a[0][1] = a[0][0] * m.a[0][1] + a[0][1] * m.a[1][1] + a[0][2] * m.a[2][1];
	res.a[1][1] = a[1][0] * m.a[0][1] + a[1][1] * m.a[1][1] + a[1][2] * m.a[2][1];
	res.a[2][1] = a[2][0] * m.a[0][1] + a[2][1] * m.a[1][1] + a[2][2] * m.a[2][1];

	// Third column
	res.a[0][2] = a[0][0] * m.a[0][2] + a[0][1] * m.a[1][2] + a[0][2] * m.a[2][2];
	res.a[1][2] = a[1][0] * m.a[0][2] + a[1][1] * m.a[1][2] + a[1][2] * m.a[2][2];
	res.a[2][2] = a[2][0] * m.a[0][2] + a[2][1] * m.a[1][2] + a[2][2] * m.a[2][2];

	return res;
}
